"""
Object-oriented bindings for
[CIVET](http://www.bic.mni.mcgill.ca/ServicesSoftware/CIVET).

See [https://pypi.org/project/pycivet/](https://pypi.org/project/pycivet/)
for a description and examples.
"""

import civet.extraction.starting_models as starting_models

__docformat__ = 'numpy'

__all__ = [
    'extraction',
    'minc',
    'obj',
    'xfm',
    'starting_models',
    'globals',
    'bases',
    'memoization',
    'abstract_data',
    'memoization',
    'shells'
]
